from __future__ import annotations

from collections.abc import (
    Awaitable,
    Callable,
    Collection,
    Iterable,
    Mapping,
    Sequence,
)
from contextlib import asynccontextmanager, contextmanager
from functools import wraps
from itertools import filterfalse, product, repeat, starmap
from pathlib import Path, PurePath
from shutil import move
from tempfile import NamedTemporaryFile
from typing import TypeVar

import sqlalchemy as sa
from attrs import evolve
from loguru import logger
from typing_extensions import Concatenate, Never, ParamSpec
from yarl import URL

from . import pkg_db, pkg_models
from . import results as R
from .common import Defn, Strategy
from .http import make_defn_progress_ctx
from .manager_ctx import ManagerCtx
from .resolvers import HeadersIntent
from .utils import (
    bucketise,
    chain_dict,
    file_uri_to_path,
    find_addon_zip_tocs,
    gather,
    is_file_uri,
    make_zip_member_filter_fn,
    run_in_thread,
    shasum,
    time_op,
    trash,
    uniq,
)

_P = ParamSpec('_P')
_T = TypeVar('_T')

_AsyncNamedTemporaryFile = run_in_thread(NamedTemporaryFile)
_move_async = run_in_thread(move)

_MUTATE_PKGS_LOCK = '_MUTATE_PKGS_'
_DOWNLOAD_PKG_LOCK = '_DOWNLOAD_PKG_'


def bucketise_results(
    values: Iterable[tuple[Defn, R.AnyResult[_T]]],
) -> tuple[Mapping[Defn, _T], Mapping[Defn, R.AnyResult[Never]]]:
    ts: dict[Defn, _T] = {}
    errors: dict[Defn, R.AnyResult[Never]] = {}

    for defn, value in values:
        if isinstance(value, (R.ManagerError, R.InternalError)):
            errors[defn] = value
        else:
            ts[defn] = value

    return ts, errors


@asynccontextmanager
async def _open_temp_writer_async():
    fh = await _AsyncNamedTemporaryFile(delete=False)
    path = Path(fh.name)
    try:
        yield (path, run_in_thread(fh.write))
    except BaseException:
        await run_in_thread(fh.close)()
        await run_in_thread(path.unlink)()
        raise
    else:
        await run_in_thread(fh.close)()


@contextmanager
def _open_pkg_archive(path: PurePath):
    from zipfile import ZipFile

    with ZipFile(path) as archive:

        def extract(parent: Path) -> None:
            archive.extractall(parent, members=filter(make_zip_member_filter_fn(base_dirs), names))

        names = archive.namelist()
        base_dirs = {h for _, h in find_addon_zip_tocs(names)}
        yield (base_dirs, extract)


@object.__new__
class _DummyResolver:
    async def resolve(self, defns: Sequence[Defn]) -> dict[Defn, R.AnyResult[pkg_models.Pkg]]:
        return dict.fromkeys(defns, R.PkgSourceInvalid())

    async def get_changelog(self, uri: URL) -> str:
        raise R.PkgSourceInvalid


def _with_lock(lock_name: str):
    def outer(
        coro_fn: Callable[Concatenate[PkgManager, _P], Awaitable[_T]]
    ) -> Callable[Concatenate[PkgManager, _P], Awaitable[_T]]:
        @wraps(coro_fn)
        async def inner(self: PkgManager, *args: _P.args, **kwargs: _P.kwargs) -> _T:
            async with self.ctx.locks[lock_name, id(self)]:
                return await coro_fn(self, *args, **kwargs)

        return inner

    return outer


@run_in_thread
def _install_pkg(
    ctx: ManagerCtx, pkg: pkg_models.Pkg, archive: Path, replace: bool
) -> R.PkgInstalled:
    with _open_pkg_archive(archive) as (top_level_folders, extract):
        with ctx.database.connect() as connection:
            installed_conflicts = connection.execute(
                sa.select(pkg_db.pkg)
                .distinct()
                .join(pkg_db.pkg_folder)
                .where(pkg_db.pkg_folder.c.name.in_(top_level_folders))
            ).all()
        if installed_conflicts:
            raise R.PkgConflictsWithInstalled(installed_conflicts)

        if replace:
            trash(
                (ctx.config.addon_dir / f for f in top_level_folders),
                dest=ctx.config.global_config.temp_dir,
                missing_ok=True,
            )
        else:
            unreconciled_conflicts = top_level_folders & {
                f.name for f in ctx.config.addon_dir.iterdir()
            }
            if unreconciled_conflicts:
                raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

        extract(ctx.config.addon_dir)

    pkg = evolve(pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)])
    with ctx.database.begin() as transaction:
        pkg.insert(transaction)

    return R.PkgInstalled(pkg)


@run_in_thread
def _update_pkg(
    ctx: ManagerCtx, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg, archive: Path
) -> R.PkgUpdated:
    with _open_pkg_archive(archive) as (top_level_folders, extract):
        with ctx.database.connect() as connection:
            installed_conflicts = connection.execute(
                sa.select(pkg_db.pkg)
                .distinct()
                .join(pkg_db.pkg_folder)
                .where(
                    pkg_db.pkg_folder.c.pkg_source != new_pkg.source,
                    pkg_db.pkg_folder.c.pkg_id != new_pkg.id,
                    pkg_db.pkg_folder.c.name.in_(top_level_folders),
                )
            ).all()
        if installed_conflicts:
            raise R.PkgConflictsWithInstalled(installed_conflicts)

        unreconciled_conflicts = top_level_folders - {f.name for f in old_pkg.folders} & {
            f.name for f in ctx.config.addon_dir.iterdir()
        }
        if unreconciled_conflicts:
            raise R.PkgConflictsWithUnreconciled(unreconciled_conflicts)

        trash(
            (ctx.config.addon_dir / f.name for f in old_pkg.folders),
            dest=ctx.config.global_config.temp_dir,
            missing_ok=True,
        )
        extract(ctx.config.addon_dir)

    new_pkg = evolve(
        new_pkg, folders=[pkg_models.PkgFolder(name=f) for f in sorted(top_level_folders)]
    )
    with ctx.database.begin() as transaction:
        old_pkg.delete(transaction)
        new_pkg.insert(transaction)

    return R.PkgUpdated(old_pkg, new_pkg)


@run_in_thread
def _remove_pkg(ctx: ManagerCtx, pkg: pkg_models.Pkg, keep_folders: bool) -> R.PkgRemoved:
    if not keep_folders:
        trash(
            (ctx.config.addon_dir / f.name for f in pkg.folders),
            dest=ctx.config.global_config.temp_dir,
            missing_ok=True,
        )

    with ctx.database.begin() as transaction:
        pkg.delete(transaction)

    return R.PkgRemoved(pkg)


class PkgManager:
    def __init__(self, ctx: ManagerCtx) -> None:
        self.ctx = ctx

    def pair_uri(self, value: str) -> tuple[str, str] | None:
        "Attempt to extract a valid ``Defn`` source and alias from a URL."
        aliases_from_url = (
            (r.metadata.id, a)
            for r in self.ctx.resolvers.values()
            for a in (r.get_alias_from_url(URL(value)),)
            if a
        )
        return next(aliases_from_url, None)

    def check_pkg_exists(self, defn: Defn) -> bool:
        "Check that a package exists in the database."
        with self.ctx.database.connect() as connection:
            return (
                connection.execute(
                    sa.select(sa.text('1'))
                    .select_from(pkg_db.pkg)
                    .where(
                        pkg_db.pkg.c.source == defn.source,
                        (pkg_db.pkg.c.id == defn.alias)
                        | (pkg_db.pkg.c.id == defn.id)
                        | (sa.func.lower(pkg_db.pkg.c.slug) == sa.func.lower(defn.alias)),
                    )
                ).scalar()
                == 1
            )

    def get_pkg(self, defn: Defn, partial_match: bool = False) -> pkg_models.Pkg | None:
        "Retrieve a package from the database."
        with self.ctx.database.connect() as connection:
            maybe_row_mapping = (
                connection.execute(
                    sa.select(pkg_db.pkg).where(
                        pkg_db.pkg.c.source == defn.source,
                        (pkg_db.pkg.c.id == defn.alias)
                        | (pkg_db.pkg.c.id == defn.id)
                        | (sa.func.lower(pkg_db.pkg.c.slug) == sa.func.lower(defn.alias)),
                    )
                )
                .mappings()
                .one_or_none()
            )
            if maybe_row_mapping is None and partial_match:
                maybe_row_mapping = (
                    connection.execute(
                        sa.select(pkg_db.pkg)
                        .where(pkg_db.pkg.c.slug.contains(defn.alias))
                        .order_by(pkg_db.pkg.c.name)
                    )
                    .mappings()
                    .first()
                )
            if maybe_row_mapping is not None:
                return pkg_models.Pkg.from_row_mapping(connection, maybe_row_mapping)

    async def find_equivalent_pkg_defns(
        self, pkgs: Collection[pkg_models.Pkg]
    ) -> dict[pkg_models.Pkg, list[Defn]]:
        "Given a list of packages, find ``Defn``s of each package from other sources."
        from .matchers import AddonFolder

        catalogue = await self.ctx.synchronise()

        @run_in_thread
        def collect_addon_folders():
            return {
                p: frozenset(
                    a
                    for f in p.folders
                    for a in (
                        AddonFolder.from_addon_path(
                            self.ctx.config.game_flavour,
                            self.ctx.config.addon_dir / f.name,
                        ),
                    )
                    if a
                )
                for p in pkgs
            }

        async def collect_hashed_folder_defns():
            matches = [
                m
                for r in self.ctx.resolvers.values()
                for m in await r.get_folder_hash_matches(
                    [a for p, f in folders_per_pkg.items() if p.source != r.metadata.id for a in f]
                )
            ]

            defns_per_pkg: dict[pkg_models.Pkg, frozenset[Defn]] = {
                p: frozenset() for p in folders_per_pkg
            }
            for (pkg, orig_folders), (defn, matched_folders) in product(
                folders_per_pkg.items(), matches
            ):
                if orig_folders == matched_folders:
                    defns_per_pkg[pkg] |= frozenset((defn,))

            return defns_per_pkg

        def get_catalogue_defns(pkg: pkg_models.Pkg) -> frozenset[Defn]:
            entry = catalogue.keyed_entries.get((pkg.source, pkg.id))
            if entry:
                return frozenset(Defn(s.source, s.id) for s in entry.same_as)
            else:
                return frozenset()

        def get_addon_toc_defns(pkg_source: str, addon_folders: Collection[AddonFolder]):
            return frozenset(
                d
                for a in addon_folders
                for d in a.get_defns_from_toc_keys(self.ctx.resolvers.addon_toc_key_and_id_pairs)
                if d.source != pkg_source
            )

        folders_per_pkg = await collect_addon_folders()

        with time_op(lambda t: logger.debug(f'hashed folder matches found in {t:.3f}s')):
            hashed_folder_defns = await collect_hashed_folder_defns()

        return {
            p: sorted(d, key=lambda d: self.ctx.resolvers.priority_dict[d.source])
            for p in pkgs
            for d in (
                get_catalogue_defns(p)
                | get_addon_toc_defns(p.source, folders_per_pkg[p])
                | hashed_folder_defns[p],
            )
            if d
        }

    async def _resolve_deps(
        self, results: Collection[R.AnyResult[pkg_models.Pkg]]
    ) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
        """Resolve package dependencies.

        The resolver will not follow dependencies
        more than one level deep.  This is to avoid unnecessary
        complexity for something that I would never expect to
        encounter in the wild.
        """
        pkgs = [r for r in results if isinstance(r, pkg_models.Pkg)]
        dep_defns = uniq(
            filterfalse(
                {(p.source, p.id) for p in pkgs}.__contains__,
                ((p.source, d.id) for p in pkgs for d in p.deps),
            )
        )
        if not dep_defns:
            return {}

        deps = await self.resolve(list(starmap(Defn, dep_defns)))
        pretty_deps = {
            evolve(d, alias=r.slug) if isinstance(r, pkg_models.Pkg) else d: r
            for d, r in deps.items()
        }
        return pretty_deps

    async def resolve(
        self, defns: Collection[Defn], with_deps: bool = False
    ) -> Mapping[Defn, R.AnyResult[pkg_models.Pkg]]:
        "Resolve definitions into packages."
        if not defns:
            return {}

        defns_by_source = bucketise(defns, key=lambda v: v.source)
        results = await gather(
            (
                self.ctx.resolvers.get(s, _DummyResolver).resolve(b)
                for s, b in defns_by_source.items()
            ),
            R.resultify_async_exc,
        )
        results_by_defn = chain_dict(
            defns,
            R.ManagerError(),
            *(
                r.items() if isinstance(r, dict) else zip(d, repeat(r))
                for d, r in zip(defns_by_source.values(), results)
            ),
        )
        if with_deps:
            results_by_defn.update(await self._resolve_deps(results_by_defn.values()))
        return results_by_defn

    async def get_changelog(self, source: str, uri: str) -> str:
        "Retrieve a changelog from a URI."
        return await self.ctx.resolvers.get(source, _DummyResolver).get_changelog(URL(uri))

    async def _download_pkg_archive(
        self, defn: Defn, pkg: pkg_models.Pkg, *, chunk_size: int = 4096
    ):
        if is_file_uri(pkg.download_url):
            return Path(file_uri_to_path(pkg.download_url))

        async with self.ctx.locks[_DOWNLOAD_PKG_LOCK, pkg.download_url]:
            dest = self.ctx.config.global_config.cache_dir / shasum(pkg.download_url)
            if not await run_in_thread(dest.exists)():
                async with self.ctx.web_client.get(
                    pkg.download_url,
                    headers=await self.ctx.resolvers[pkg.source].make_request_headers(
                        intent=HeadersIntent.Download
                    ),
                    raise_for_status=True,
                    trace_request_ctx=make_defn_progress_ctx(self.ctx.config.profile, defn),
                ) as response, _open_temp_writer_async() as (
                    temp_path,
                    write,
                ):
                    async for chunk in response.content.iter_chunked(chunk_size):
                        await write(chunk)

                await _move_async(temp_path, dest)

            return dest

    @run_in_thread
    def _check_installed_pkg_integrity(self, pkg: pkg_models.Pkg) -> bool:
        return all((self.ctx.config.addon_dir / p.name).exists() for p in pkg.folders)

    async def _should_update_pkg(self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg) -> bool:
        return old_pkg.version != new_pkg.version or (
            not await self._check_installed_pkg_integrity(old_pkg)
        )

    @_with_lock(_MUTATE_PKGS_LOCK)
    async def install(
        self, defns: Sequence[Defn], replace: bool, dry_run: bool = False
    ) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
        "Install packages from a definition list."

        # We'll weed out installed deps from the results after resolving -
        # doing it this way isn't particularly efficient but avoids having to
        # deal with local state in ``resolve``
        resolve_results = await self.resolve(
            [d for d in defns if not self.check_pkg_exists(d)], with_deps=True
        )
        pkgs, resolve_errors = bucketise_results(resolve_results.items())
        new_pkgs = {d: p for d, p in pkgs.items() if not self.check_pkg_exists(p.to_defn())}

        results = dict.fromkeys(defns, R.PkgAlreadyInstalled()) | resolve_errors

        if dry_run:
            return results | {d: R.PkgInstalled(p, dry_run=True) for d, p in new_pkgs.items()}

        download_results = zip(
            new_pkgs,
            await gather(
                (self._download_pkg_archive(d, r) for d, r in new_pkgs.items()),
                R.resultify_async_exc,
            ),
        )
        archive_paths, download_errors = bucketise_results(download_results)

        return (
            results
            | download_errors
            | {
                d: await R.resultify_async_exc(_install_pkg(self.ctx, new_pkgs[d], a, replace))
                for d, a in archive_paths.items()
            }
        )

    @_with_lock(_MUTATE_PKGS_LOCK)
    async def update(
        self, defns: Sequence[Defn], retain_defn_strategy: bool, dry_run: bool = False
    ) -> Mapping[Defn, R.AnyResult[R.PkgInstalled | R.PkgUpdated]]:
        """Update installed packages from a definition list.

        A ``retain_defn_strategy`` value of false will instruct ``update``
        to extract the strategy from the installed package; otherwise
        the ``Defn`` strategy will be used.
        """

        defns_to_pkgs = {d: p for d in defns for p in (self.get_pkg(d),) if p}
        resolve_defns = {
            # Attach the source ID to each ``Defn`` from the
            # corresponding installed package.  Using the ID has the benefit
            # of resolving installed-but-renamed packages - the slug is
            # transient but the ID isn't
            evolve(d, id=p.id) if retain_defn_strategy else p.to_defn(): d
            for d, p in defns_to_pkgs.items()
        }
        resolve_results = await self.resolve(resolve_defns, with_deps=True)
        # Discard the reconstructed ``Defn``s
        orig_defn_resolve_results = (
            (resolve_defns.get(d, d), r) for d, r in resolve_results.items()
        )
        pkgs, resolve_errors = bucketise_results(orig_defn_resolve_results)

        updatables = {
            d: (o, n)
            for d, n in pkgs.items()
            for o in (defns_to_pkgs.get(d),)
            if not o or await self._should_update_pkg(o, n)
        }

        results = (
            dict.fromkeys(defns, R.PkgNotInstalled())
            | resolve_errors
            | {
                d: R.PkgUpToDate(is_pinned=pkgs[d].options.version_eq)
                for d in pkgs.keys() - updatables.keys()
            }
        )

        if dry_run:
            return results | {
                d: R.PkgUpdated(o, n, dry_run=True) if o else R.PkgInstalled(n, dry_run=True)
                for d, (o, n) in updatables.items()
            }

        download_results = zip(
            updatables,
            await gather(
                (self._download_pkg_archive(d, n) for d, (_, n) in updatables.items()),
                R.resultify_async_exc,
            ),
        )
        archive_paths, download_errors = bucketise_results(download_results)

        return (
            results
            | download_errors
            | {
                d: await R.resultify_async_exc(
                    _update_pkg(self.ctx, o, n, a) if o else _install_pkg(self.ctx, n, a, False)
                )
                for d, a in archive_paths.items()
                for o, n in (updatables[d],)
            }
        )

    @_with_lock(_MUTATE_PKGS_LOCK)
    async def remove(
        self, defns: Sequence[Defn], keep_folders: bool
    ) -> Mapping[Defn, R.AnyResult[R.PkgRemoved]]:
        "Remove packages by their definition."
        return {
            d: (
                await R.resultify_async_exc(_remove_pkg(self.ctx, p, keep_folders))
                if p
                else R.PkgNotInstalled()
            )
            for d in defns
            for p in (self.get_pkg(d),)
        }

    @_with_lock(_MUTATE_PKGS_LOCK)
    async def pin(self, defns: Sequence[Defn]) -> Mapping[Defn, R.AnyResult[R.PkgInstalled]]:
        """Pin and unpin installed packages.

        instawow does not have true pinning.  This flips ``Strategy.VersionEq``
        on for installed packages from sources that support it.
        The net effect is the same as if the package
        had been reinstalled with the ``VersionEq`` strategy.
        """

        @run_in_thread
        def pin(defn: Defn) -> R.PkgInstalled:
            resolver = self.ctx.resolvers.get(defn.source)
            if resolver is None:
                raise R.PkgSourceInvalid

            if Strategy.VersionEq not in resolver.metadata.strategies:
                raise R.PkgStrategiesUnsupported({Strategy.VersionEq})

            pkg = self.get_pkg(defn)
            if not pkg:
                raise R.PkgNotInstalled

            version = defn.strategies.version_eq
            if version and pkg.version != version:
                R.PkgFilesNotMatching(defn.strategies)

            with self.ctx.database.begin() as transaction:
                transaction.execute(
                    sa.update(pkg_db.pkg_options)
                    .filter_by(pkg_source=pkg.source, pkg_id=pkg.id)
                    .values(version_eq=version is not None)
                )

            new_pkg = evolve(pkg, options=evolve(pkg.options, version_eq=version is not None))
            return R.PkgInstalled(new_pkg)

        return {d: await R.resultify_async_exc(pin(d)) for d in defns}



"""Renovate ``pkg_options``.

Revision ID: 98716a7301f8
Revises: e13430219249
Create Date: 2022-10-23 16:18:51.169154

"""
from __future__ import annotations

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '98716a7301f8'
down_revision = 'e13430219249'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('pkg_options', schema=None) as batch_op:
        batch_op.add_column(sa.Column('any_flavour', sa.Boolean()))
        batch_op.add_column(sa.Column('any_release_type', sa.Boolean()))
        batch_op.add_column(sa.Column('version_eq', sa.Boolean()))

    pkg_options = sa.table(
        'pkg_options',
        sa.column('any_flavour', sa.Boolean),
        sa.column('any_release_type', sa.Boolean),
        sa.column('version_eq', sa.Boolean),
        sa.column('strategy', sa.String),
    )

    conn = op.get_bind()
    conn.execute(
        sa.update(pkg_options)
        .where(pkg_options.c.strategy == sa.bindparam('strategy_'))
        .values(
            any_flavour=sa.bindparam('any_flavour'),
            any_release_type=sa.bindparam('any_release_type'),
            version_eq=sa.bindparam('version_eq'),
        ),
        [
            {
                'strategy_': 'default',
                'any_flavour': False,
                'any_release_type': False,
                'version_eq': False,
            },
            {
                'strategy_': 'latest',
                'any_flavour': False,
                'any_release_type': True,
                'version_eq': False,
            },
            {
                'strategy_': 'any_flavour',
                'any_flavour': True,
                'any_release_type': False,
                'version_eq': False,
            },
            {
                'strategy_': 'version',
                'any_flavour': False,
                'any_release_type': False,
                'version_eq': True,
            },
        ],
    )

    with op.batch_alter_table('pkg_options', schema=None) as batch_op:
        batch_op.alter_column('any_flavour', nullable=False)
        batch_op.alter_column('any_release_type', nullable=False)
        batch_op.alter_column('version_eq', nullable=False)
        batch_op.drop_column('strategy')


def downgrade():
    with op.batch_alter_table('pkg_options', schema=None) as batch_op:
        batch_op.add_column(sa.Column('strategy', sa.String(), nullable=False))
        batch_op.drop_column('version_eq')
        batch_op.drop_column('any_release_type')
        batch_op.drop_column('any_flavour')



from __future__ import annotations

import sys
from pathlib import Path

from loguru import logger


def _patch_loguru_enqueue():
    import queue
    import threading
    from types import SimpleNamespace

    import loguru._handler

    # On some systems when instantiating multiprocessing constructs Python
    # starts the multiprocessing resource monitor.  This is done in a subprocess
    # with ``sys.executable`` as the first argument.  In briefcase
    # this points to the same executable that starts the app - there
    # isn't a separate Python executable.  So with every new resource
    # monitor that spawns so does a new copy of the app, ad infinitum.
    # Even when not using briefcase spawning a subprocess slows down start-up,
    # not least because it imports a second copy of the ``site`` module.
    # We replace these multiprocessing constructs with their threading
    # equivalents in loguru since loguru itself does not spawn a subprocesses
    # but creates a separate thread for its "enqueued" logger and we don't
    # use multiprocessing in instawow.
    # This will definitely not come back to bite us.
    loguru._handler.multiprocessing = SimpleNamespace(  # pyright: ignore[reportPrivateImportUsage]
        SimpleQueue=queue.Queue, Event=threading.Event, Lock=threading.Lock
    )


def _intercept_logging_module_calls(log_level: str):  # pragma: no cover
    import logging

    logging_filename = getattr(logging, '__file__', None)

    class InterceptHandler(logging.Handler):
        def emit(self, record: logging.LogRecord) -> None:
            # Get the corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where the logged message originated
            depth = 6
            frame = sys._getframe(depth)  # pyright: ignore[reportPrivateUsage]
            while frame and frame.f_code.co_filename == logging_filename:
                frame = frame.f_back
                depth += 1

            logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())

    logging.basicConfig(handlers=[InterceptHandler()], level=log_level, force=True)


def setup_logging(
    logging_dir: Path, log_to_stderr: bool, debug: bool, intercept_logging_module_calls: bool
) -> None:
    _patch_loguru_enqueue()

    log_level = 'DEBUG' if debug else 'INFO'

    if intercept_logging_module_calls:
        _intercept_logging_module_calls(log_level)

    handlers = [
        {
            'level': log_level,
            'enqueue': True,
            'sink': logging_dir / 'error.log',
            'rotation': '5 MB',
            'retention': 5,  # Number of log files to keep
        },
    ]
    if log_to_stderr:
        handlers += [
            {
                'level': log_level,
                'enqueue': True,
                'sink': sys.stderr,
                'format': (
                    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
                    '<level>{level: <8}</level> | '
                    '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>\n'
                    '  <level>{message}</level>'
                ),
            }
        ]

    logger.configure(handlers=handlers)



from __future__ import annotations

import json
import os
import sys
import typing
from collections.abc import Callable, Iterable, Mapping, Sized
from functools import lru_cache, partial
from pathlib import Path
from tempfile import gettempdir
from typing import Any, TypedDict, TypeVar

import cattrs.preconf.json
from attrs import Attribute, field, fields, frozen, has
from cattrs import AttributeValidationNote, Converter
from cattrs.gen import make_dict_unstructure_fn, override
from typing_extensions import Self

from .common import Flavour
from .utils import add_exc_note, trash

_T = TypeVar('_T')

_MISSING = object()

_BOTTOM_DIR_NAME = 'instawow'


class SecretStr(str):
    pass


def _expand_path(value: Path):
    return Path(os.path.abspath(os.path.expanduser(value)))


def _is_writable_dir(value: Path):
    return value.is_dir() and os.access(value, os.W_OK)


def _ensure_dirs(dirs: Iterable[Path]):
    for dir_ in dirs:
        dir_.mkdir(exist_ok=True, parents=True)


def _enrich_validator_exc(validator: Callable[[object, Attribute[_T], _T], None]):
    def wrapper(model: object, attr: Attribute[_T], value: _T):
        try:
            validator(model, attr, value)
        except BaseException as exc:
            note = f'Structuring class {model.__class__.__name__} @ attribute {attr.name}'
            add_exc_note(exc, AttributeValidationNote(note, attr.name, attr.type))
            raise

    return wrapper


@_enrich_validator_exc
def _validate_path_is_writable_dir(_model: object, _attr: Attribute[Path], value: Path):
    if not _is_writable_dir(value):
        raise ValueError(f'"{value}" is not a writable directory')


def _make_validate_min_length(min_length: int):
    @_enrich_validator_exc
    def _validate_min_length(_model: object, _attr: Attribute[Sized], value: Sized):
        if len(value) < min_length:
            raise ValueError(f'Value must have a minimum length of {min_length}')

    return _validate_min_length


def _encode_config_for_display(config: object):
    converter = _make_display_converter()
    return json.dumps(
        converter.unstructure(config),
        indent=2,
    )


def _write_config(config: object, config_path: Path):
    converter = _make_write_converter()
    config_path.write_text(
        json.dumps(converter.unstructure(config), indent=2),
        encoding='utf-8',
    )


def _read_config(config_path: Path, missing_ok: bool = False) -> dict[str, Any]:
    try:
        return json.loads(config_path.read_bytes())
    except FileNotFoundError:
        if missing_ok:
            return {}
        raise


def _compute_var(field_: Attribute[object], default: object):
    if not field_.metadata.get('from_env'):
        return default

    value = os.environ.get(f'instawow_{field_.name}'.upper())
    if value is None:
        return default
    elif field_.metadata.get('as_json'):
        return json.loads(value)
    else:
        return value


def _read_env_vars(config_cls: Any, values: Mapping[str, object]):
    return {
        f.name: v
        for f in fields(config_cls)
        for v in (_compute_var(f, values.get(f.name, _MISSING)),)
        if v is not _MISSING
    }


def _make_attrs_instance_hook_factory(converter: Converter, type_: type):
    "Allow passing in a structured attrs instance to ``structure``."
    structure = converter.gen_structure_attrs_fromdict(type_)

    def structure_wrapper(value: Mapping[str, Any], type_: type):
        return value if isinstance(value, type_) else structure(value, type_)

    return structure_wrapper


def make_config_converter():
    converter = Converter()
    cattrs.preconf.json.configure_converter(converter)
    converter.register_structure_hook(Path, lambda v, _: Path(v))
    converter.register_unstructure_hook(Path, str)
    return converter


@lru_cache(1)
def _make_display_converter():
    def convert_secret_string(value: str):
        return '**********'

    def convert_global_config(global_config: GlobalConfig):
        return converter.unstructure_attrs_asdict(global_config) | {
            'profiles': global_config.list_profiles()
        }

    converter = make_config_converter()
    converter.register_unstructure_hook(GlobalConfig, convert_global_config)
    converter.register_unstructure_hook(SecretStr, convert_secret_string)
    return converter


@lru_cache(1)
def _make_write_converter():
    converter = make_config_converter()
    for config_cls in [GlobalConfig, Config]:
        converter.register_unstructure_hook(
            config_cls,
            make_dict_unstructure_fn(
                config_cls,
                converter,
                **{  # pyright: ignore[reportGeneralTypeIssues]  # See microsoft/pyright#5255
                    f.name: override(omit=True)
                    for f in fields(config_cls)
                    if not f.metadata.get('write_on_disk')
                },
            ),
        )
    return converter


def _get_default_config_dir():
    config_parent_dir = os.environ.get('XDG_CONFIG_HOME')

    if not config_parent_dir:
        if sys.platform == 'darwin':
            config_parent_dir = Path.home() / 'Library' / 'Application Support'
        elif sys.platform == 'win32':
            config_parent_dir = os.environ.get('APPDATA')

    if not config_parent_dir:
        config_parent_dir = Path.home() / '.config'

    return Path(config_parent_dir, _BOTTOM_DIR_NAME)


def _get_default_temp_dir():
    return Path(gettempdir(), _BOTTOM_DIR_NAME)


def _get_default_state_dir():
    state_parent_dir = os.environ.get('XDG_STATE_HOME')

    if not state_parent_dir and sys.platform not in {'darwin', 'win32'}:
        state_parent_dir = Path.home() / '.local' / 'state'

    if not state_parent_dir:
        return _get_default_config_dir()

    return Path(state_parent_dir, _BOTTOM_DIR_NAME)


class _ConfigMetadata(TypedDict, total=False):
    from_env: bool
    as_json: bool
    write_on_disk: bool


_AccessToken = typing.Union[SecretStr, None]


@frozen
class _AccessTokens:
    cfcore: _AccessToken = None
    github: _AccessToken = None
    wago: _AccessToken = None
    wago_addons: _AccessToken = None


@frozen
class GlobalConfig:
    config_dir: Path = field(
        factory=_get_default_config_dir,
        converter=_expand_path,
        metadata=_ConfigMetadata(from_env=True),
    )
    temp_dir: Path = field(
        factory=_get_default_temp_dir,
        converter=_expand_path,
        metadata=_ConfigMetadata(from_env=True),
    )
    state_dir: Path = field(
        factory=_get_default_state_dir,
        converter=_expand_path,
        metadata=_ConfigMetadata(from_env=True),
    )
    auto_update_check: bool = field(
        default=True,
        metadata=_ConfigMetadata(from_env=True, as_json=True, write_on_disk=True),
    )
    access_tokens: _AccessTokens = field(
        default=_AccessTokens(),
        metadata=_ConfigMetadata(from_env=True, as_json=True, write_on_disk=True),
    )

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = _read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls) -> Self:
        env_only_config = cls.from_values(env=True)
        config_values = _read_config(env_only_config.config_file, missing_ok=True)
        return cls.from_values(config_values, env=True) if config_values else env_only_config

    def list_profiles(self) -> list[str]:
        "Get the names of the profiles contained in ``config_dir``."
        profiles = [c.parent.name for c in self.profiles_config_dir.glob('*/config.json')]
        return profiles

    def ensure_dirs(self) -> Self:
        _ensure_dirs(
            [
                self.config_dir,
                self.temp_dir,
                self.state_dir,
                self.cache_dir,
            ]
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        _write_config(self, self.config_file)
        return self

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / 'cache'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def profiles_config_dir(self) -> Path:
        return self.config_dir / 'profiles'

    @property
    def profiles_state_dir(self) -> Path:
        return self.state_dir / 'profiles'


@frozen
class Config:
    global_config: GlobalConfig
    profile: str = field(
        converter=str.strip,
        validator=_make_validate_min_length(1),
        metadata=_ConfigMetadata(from_env=True, write_on_disk=True),
    )
    addon_dir: Path = field(
        converter=_expand_path,
        validator=_validate_path_is_writable_dir,
        metadata=_ConfigMetadata(from_env=True, write_on_disk=True),
    )
    game_flavour: Flavour = field(metadata=_ConfigMetadata(from_env=True, write_on_disk=True))

    @classmethod
    def make_dummy_config(cls, **values: object) -> Self:
        return object.__new__(type(f'Dummy{cls.__name__}', (cls,), values))

    @classmethod
    def from_values(cls, values: Mapping[str, object] = {}, *, env: bool = False) -> Self:
        if env:
            values = _read_env_vars(cls, values)
        return config_converter.structure(values, cls)

    @classmethod
    def read(cls, global_config: GlobalConfig, profile: str) -> Self:
        dummy_config = cls.make_dummy_config(global_config=global_config, profile=profile)
        return cls.from_values(
            {**_read_config(dummy_config.config_file), 'global_config': global_config},
            env=True,
        )

    def encode_for_display(self) -> str:
        return _encode_config_for_display(self)

    def ensure_dirs(self) -> Self:
        _ensure_dirs(
            [
                self.config_dir,
                self.state_dir,
                self.logging_dir,
                self.plugins_dir,
            ]
        )
        return self

    def write(self) -> Self:
        self.ensure_dirs()
        _write_config(self, self.config_file)
        return self

    def delete(self) -> None:
        trash((self.config_dir,), dest=self.global_config.temp_dir, missing_ok=True)

    @property
    def config_dir(self) -> Path:
        return self.global_config.profiles_config_dir / self.profile

    @property
    def state_dir(self) -> Path:
        return self.global_config.profiles_state_dir / self.profile

    @property
    def logging_dir(self) -> Path:
        return self.state_dir / 'logs'

    @property
    def plugins_dir(self) -> Path:
        return self.state_dir / 'plugins'

    @property
    def config_file(self) -> Path:
        return self.config_dir / 'config.json'

    @property
    def db_uri(self) -> str:
        return f"sqlite:///{self.config_dir / 'db.sqlite'}"


config_converter = make_config_converter()
config_converter.register_structure_hook_factory(
    has, partial(_make_attrs_instance_hook_factory, config_converter)
)
